//-< JAVAGO.H >------------------------------------------------------*--------*
// JavaGO                     Version 1.02       (c) 1998  GARRET    *     ?  *
// (Java Global Optimizer)                                           *   /\|  *
//                                                                   *  /  \  *
//                          Created:     10-Oct-98    K.A. Knizhnik  * / [] \ *
//                          Last update: 12-Nov-98    K.A. Knizhnik  * GARRET *
//-------------------------------------------------------------------*--------*

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include <stdarg.h>
#include <ctype.h>
#ifdef _WIN32
#include <windows.h>
#else
#include <dirent.h>
#endif

#ifndef MAX_PATH
#define MAX_PATH 1024
#endif

#include "javago.h"

static int const vbm_instruction_length[] = {
#define JAVA_INSN(code, mnem, len) len,
#include "javago.d"
0
};

int inline_threshold = 64;
int n_inline_calls;
int n_nonvirtual_calls;

bool not_inline_constructor;
bool not_inline_virtual;
bool not_use_invokespecial;

int min_call_count;
char* profile_name;

unsigned string_hash_function(byte* p, int len) { 
    unsigned h = 0, g;
    while (--len >= 0) { 
	h = (h << 4) + *p++;
	if ((g = h & 0xF0000000) != 0) { 
	    h ^= g >> 24;
	}
	h &= ~g;
    }
    return h;
}

byte* generation_context::extend_constant_buffer(int size)
{
    if (constant_ptr + size > constant_buffer_end) {
	int old_size = constant_ptr - constant_buffer;
	int new_size = size > old_size ? size*2 : old_size*2;
	byte* new_buffer = new byte[new_size];
	memcpy(new_buffer, constant_buffer, old_size);
	delete[] constant_buffer;
	constant_buffer = new_buffer;
	constant_buffer_end = new_buffer + new_size;
	constant_ptr = new_buffer + old_size;
    }
    byte* ptr = constant_ptr;
    constant_ptr += size;
    return ptr;
}

byte* generation_context::extend_exception_buffer(int size)
{
    if (exception_ptr + size > exception_buffer_end) {
	int old_size = exception_ptr - exception_buffer;
	int new_size = size > old_size ? size*2 : old_size*2;
	byte* new_buffer = new byte[new_size];
	memcpy(new_buffer, exception_buffer, old_size);
	delete[] exception_buffer;
	exception_buffer = new_buffer;
	exception_buffer_end = new_buffer + new_size;
	exception_ptr = new_buffer + old_size;
    }
    byte* ptr = exception_ptr;
    exception_ptr += size;
    return ptr;
}

void generation_context::reset_for_new_method()
{
    exception_ptr = exception_buffer;
    code_ptr = code_buffer;
    mapped_variables  = 0;
    max_locals = 0;
    max_stack = 0;
}

void generation_context::reset_for_new_class(class_file* cls)
{
    this_class = cls;
    constant_ptr = constant_buffer;
    constant_pool_count = cls->constant_pool_count; 
}


generation_context::generation_context() { 
    const int init_exception_buffer_size = 1024;
    const int init_constant_buffer_size = 16*1024;
    exception_buffer = new byte[init_exception_buffer_size];
    constant_buffer = new byte[init_constant_buffer_size];
    exception_buffer_end = exception_buffer + init_exception_buffer_size;
    constant_buffer_end = constant_buffer + init_constant_buffer_size;
}





class_file* class_file::class_list;
class_file* class_file::hash_table[CLASS_HASH_TABLE_SIZE];
generation_context class_file::ctx;
relocation_record* relocation_record::free;

class_file* parse_class_file(const char* name, int len, byte* buf, bool zip) 
{ 
    unsigned magic = unpack4(buf); 
    if (magic != 0xCAFEBABE) { 
	return NULL;
    }
    return new class_file(name, len, buf, zip);
}


void load_file(const char* file_name, int recursive = 0)
{
#ifdef _WIN32
    HANDLE dir;
    char dir_path[MAX_PATH];
    WIN32_FIND_DATA file_data;
    if (recursive != 0) { 
	sprintf(dir_path, "%s\\*", file_name);
	if ((dir=FindFirstFile(dir_path, &file_data)) != INVALID_HANDLE_VALUE)
	{ 
	    file_name = dir_path; 
	} 
    } else { 
	if (strcmp(file_name, "..") == 0 || strcmp(file_name, ".") == 0) { 
	    load_file(file_name, 1);
	    return;
	}
	if ((dir=FindFirstFile(file_name, &file_data)) == INVALID_HANDLE_VALUE)
	{
	    fprintf(stderr, "Failed to open file '%s'\n", file_name);
	    return;
	}
    }
    if (dir != INVALID_HANDLE_VALUE) {
	do {
	    if (!recursive || *file_data.cFileName != '.') { 
		char file_path[MAX_PATH];
		char* file_dir = strrchr(file_name, '\\');
		char* file_name_with_path;
		if (file_dir != NULL) { 
		    int dir_len = file_dir - file_name + 1;
		    memcpy(file_path, file_name, dir_len);
		    strcpy(file_path+dir_len, file_data.cFileName);
		    file_name_with_path = file_path;
		} else { 
		    file_name_with_path = file_data.cFileName;
		}
		load_file(file_name_with_path, recursive+1);
	    }
	} while (FindNextFile(dir, &file_data));
	CloseHandle(dir);
	return;
    }
#else
    DIR* dir = opendir(file_name);
    if (dir != NULL) { 
	struct dirent* entry;
	while ((entry = readdir(dir)) != NULL) { 
	    if (*entry->d_name != '.') { 
		char full_name[MAX_PATH];
		sprintf(full_name, "%s/%s", file_name, entry->d_name);
		load_file(full_name, 2);
	    }
	} 
	closedir(dir);
	return;
    }
#endif
    if (recursive >= 2) { 
	int len = strlen(file_name); 
	if (len < 6 || (strcmp(file_name + len - 6, ".class") != 0 &&
			strcmp(file_name + len - 6, ".CLASS") != 0))
	{
	    return;
	}
    }
		
    FILE* f = fopen(file_name, "rb");
    if (f == NULL) { 
	fprintf(stderr, "Failed to open file '%s'\n", file_name);
	return;
    }

    fseek(f, 0, SEEK_END);
    size_t file_size = ftell(f);
    fseek(f, 0, SEEK_SET);
    byte* buffer = new byte[file_size];
    if (fread(buffer, 1, file_size, f) != file_size) { 
	fprintf(stderr, "Failed to read file '%s'\n", file_name);
	fclose(f);
	return;
    }

    const char* suf = file_name + strlen(file_name) - 4;

    if (suf > file_name 
	&& (strcmp(suf, ".zip") == 0 || strcmp(suf, ".ZIP") == 0 ||
	    strcmp(suf, ".jar") == 0 || strcmp(suf, ".JAR") == 0)) 
    { 
	// extract files from ZIP archive
	if (file_size < ECREC_SIZE+4) { 
	    fprintf(stderr, "Bad format of ZIP file '%s'\n", file_name);
	    return;
	}
	byte* hdr = buffer + file_size - (ECREC_SIZE+4);
	int count = unpack2_le(hdr + TOTAL_ENTRIES_CENTRAL_DIR);
	int dir_size = unpack4_le(hdr + SIZE_CENTRAL_DIRECTORY);
	byte* directory = hdr - dir_size;
	byte* dp = directory;

	while (--count >= 0) {
	    int uncompressed_size = unpack4_le(dp+4+C_UNCOMPRESSED_SIZE);
	    int filename_length = unpack2_le(dp+4+C_FILENAME_LENGTH);
	    int cextra_length = unpack2_le(dp+4+C_EXTRA_FIELD_LENGTH);

	    if ((dp-directory)+filename_length+CREC_SIZE+4 > dir_size) { 
		fprintf(stderr, "Bad format of ZIP file '%s'\n", file_name);
		break;
	    }
		
	    int local_header_offset = 
		unpack4_le(dp+4+C_RELATIVE_OFFSET_LOCAL_HEADER);
	    byte* local_header = buffer + local_header_offset;

	    if (memcmp(local_header+1, LOCAL_HDR_SIG, 3) != 0) {
		fprintf(stderr, "Bad format of ZIP file '%s'\n", file_name);
		break;
	    }
		
	    int file_start = local_header_offset + (LREC_SIZE+4) 
	        + unpack2_le(local_header+4+L_FILENAME_LENGTH)
		+ unpack2_le(local_header+4+L_EXTRA_FIELD_LENGTH);

	    int filename_offset = CREC_SIZE+4;

	    if (uncompressed_size != 0) { 
		parse_class_file((char*)dp+filename_offset,
				 filename_length, buffer+file_start, true);
	    }
	    dp += filename_offset + filename_length + cextra_length; 
	}
    } else { 
	parse_class_file(file_name, strlen(file_name), buffer, false);
    }
    fclose(f);
}

class_file::class_file(const char* name, int name_len, byte* fp, bool zip)
{
    int i;

    buffer = fp;
    archived = zip;
    final = true;
    file_name = new char[name_len+1];
    sprintf(file_name, "%.*s", name_len, name);

    constants = fp += 8;
    constant_pool_count = unpack2(fp);
    fp += 2;
    constant_pool = new constant*[constant_pool_count];
    for (i = 1; i < constant_pool_count; i++) { 
	constant* c = NULL;
	switch (*fp) { 
	  case c_utf8:
	    c = new const_utf8(fp);
	    break;
	  case c_integer:
	  case c_float:
	    c = new const_literal(fp);
	    break;
	  case c_long:
	  case c_double:
	    constant_pool[i++] = c = new const_long_literal(fp);
	    break;
	  case c_class:
	    c = new const_class(fp);
	    break;
	  case c_string:
	    c = new const_string(fp);
	    break;
	  case c_field_ref:
	  case c_method_ref:
	  case c_interface_method_ref:
	    c = new const_ref(fp);
	    break;
	  case c_name_and_type:
	    c = new const_name_and_type(fp);
	    break;
	}
	assert(c != NULL);
	fp += c->length();
	constant_pool[i] = c;
    }	    
    constant_mapping = new int[constant_pool_count];
    mapping_for_class = NULL;

    attributes = fp;
    // skip class attribute
    fp += 2;
    this_class = unpack2(fp);
    fp += 2;
    int super_class = unpack2(fp);
    fp += 2;
    int interfaces_count = unpack2(fp);
    fp += 2 + 2*interfaces_count;
    
    class_name = &((const_utf8*)constant_pool
		   [((const_class*)constant_pool[this_class])->name])->str;
    super_class_name = NULL;
    if (super_class != 0) { 
	super_class_name = 
	    &((const_utf8*)constant_pool
	      [((const_class*)constant_pool[super_class])->name])->str;
    }

    next = class_list;
    class_list = this;

    unsigned h = class_name->hash() % CLASS_HASH_TABLE_SIZE;
    collision_chain = hash_table[h];
    hash_table[h] = this;

    int fields_count = unpack2(fp); fp += 2;    
    while (--fields_count >= 0) {
	int attr_count = unpack2(fp+6); fp += 8;
	while (--attr_count >= 0) { 
	    int attr_len = unpack4(fp+2); 
	    fp += 6 + attr_len;
	}
    }

    int methods_count = unpack2(fp); fp += 2;
    methods = fp;
    method_list = NULL;
    class_method** mpp = &method_list;

    while (--methods_count >= 0) { 
	int access_flags = unpack2(fp); fp += 2;
	int name_index = unpack2(fp); fp += 2;
	int desc_index = unpack2(fp); fp += 2;
	int attr_count = unpack2(fp); fp += 2;

	class_method* method = 
	    new class_method(((const_utf8*)constant_pool[name_index])->str,
			     ((const_utf8*)constant_pool[desc_index])->str,
			     access_flags, this);
	*mpp = method;
	mpp = &method->next;

	while (--attr_count >= 0) { 
	    int attr_name = unpack2(fp);
	    int attr_len = unpack4(fp+2); 
	    const_utf8* attr = (const_utf8*)constant_pool[attr_name];
	    if (attr->str == "Code") { 
		fp += 6;
		method->max_stack = unpack2(fp); fp += 2; 
		method->max_locals = unpack2(fp); fp += 2;
 		method->code_length = unpack4(fp); fp += 4;
		method->code = fp; fp += method->code_length;

		method->exception_table_length = unpack2(fp); fp += 2;
		method->exception_table = fp; 
		fp += method->exception_table_length*8;

		int method_attr_count = unpack2(fp); fp += 2;	

		while (--method_attr_count >= 0) { 
		    int mth_attr_name = unpack2(fp); fp += 2;
		    int mth_attr_len = unpack4(fp); fp += 4;
		    const_utf8* attr = 
			(const_utf8*)constant_pool[mth_attr_name];
		    if (attr->str == "LineNumberTable") {
			method->line_table_size = unpack2(fp)*4 + 8;   
		    } else if (attr->str == "LocalVariableTable") { 
			method->local_var_table_size = unpack2(fp)*10 + 8; 
		    }
		    fp += mth_attr_len;
		}
	    } else { 
		fp += attr_len + 6;
	    }
	}
    }
    *mpp = NULL;
}

class_method* class_file::find_method(utf_string const& name, 
				      utf_string const& desc)
{
	for (class_method* mth = method_list; mth != NULL; mth = mth->next) { 
		if (mth->name == name && mth->desc == desc) {
			return mth;
		}
	}
	return NULL;
}

class_file* class_file::find(utf_string const& name) 
{
    for (class_file* cf = hash_table[name.hash() % CLASS_HASH_TABLE_SIZE];
	 cf != NULL; 
	 cf = cf->collision_chain)
    {
	if (name == *cf->class_name) { 
	    return cf;
	}
    }
    return NULL;
}


void class_file::optimize() { 
    class_file* cf;

    for (cf = class_list; cf != NULL; cf = cf->next) { 
	if (cf->super_class_name != NULL) {
	    class_file* base = find(*cf->super_class_name);
	    if (base != NULL) { 
		base->final = false;
	    }
	} else { 
	    cf->final = false;
	}
    }

    if (profile_name != NULL) { 
	FILE* profile = fopen(profile_name, "r");
	if (profile != NULL) { 
	    const int max_profile_line_length = 1024; 
	    char buf[max_profile_line_length];
	    char method[max_profile_line_length];
	    int count;

	    fgets(buf, sizeof buf, profile); // skip "count callee caller time"

	    while (fgets(buf, sizeof buf, profile) != NULL &&
		   sscanf(buf, "%d%s", &count, method) == 2)
	    {
		char *p, *mth_name = NULL, *mth_desc = NULL;
		for (p = method; *p != '\0'; p++) { 
		    if (*p == '.') { 
			mth_name = p + 1;
		    } else if (*p == '(') { 
			mth_desc = p;
		    }
		}
		if (mth_name != NULL && mth_desc != NULL) { 
		    utf_string class_name(mth_name-method-1, (byte*)method);
		    utf_string method_name(mth_desc-mth_name, (byte*)mth_name);
		    utf_string method_desc(p-mth_desc, (byte*)mth_desc);
		    
		    class_file* cls = class_file::find(class_name);
		    if (cls != NULL) { 
			class_method* mth = 
			    cls->find_method(method_name, method_desc);
			mth->call_count += count;
		    }
		}
	    }
	    fclose(profile);
	} else { 
	    fprintf(stderr, "Failed to open profile file '%s'\n",profile_name);
	}
    }
		   
    for (cf = class_list; cf != NULL; cf = cf->next) { 
	cf->optimize_file();
    }
}
		


void class_file::optimize_file()
{
    byte* fp = constants;
    class_method* mth;
    byte buf[8];

    if (archived) { 
	return;
    }

    FILE* f = fopen(file_name, "wb");
    if (f == NULL) { 
	fprintf(stderr, "Failed to open file '%s' for writing\n", file_name);
	return;
    }
    fwrite(buffer, constants - buffer, 1, f);

    ctx.reset_for_new_class(this);

    for (mth = method_list; mth != NULL; mth = mth->next) { 
	if (!(mth->attr & (f_native|f_abstract))) { 
	    ctx.reset_for_new_method();
	    mth->optimize(ctx, false);
	}
    }

    pack2(buf, ctx.constant_pool_count);
    fwrite(buf, 2, 1, f); // new size of constant pool
    // old part of constant pool
    fwrite(constants+2, attributes-constants-2, 1, f); 
    // appended part of constant pool
    fwrite(ctx.constant_buffer, ctx.constant_ptr - ctx.constant_buffer, 1, f);
    fwrite(attributes, methods-attributes, 1, f);

    for (fp = methods, mth = method_list; mth != NULL; mth = mth->next) { 
	fwrite(fp, 8, 1, f);
	int attr_count = unpack2(fp+6); fp += 8;
	
	while (--attr_count >= 0) { 
	    int attr_name = unpack2(fp);
	    int attr_len = unpack4(fp+2); 
	    const_utf8* attr = (const_utf8*)constant_pool[attr_name];

	    if (attr->str == "Code") { 
		pack4(fp+2, attr_len + mth->new_code_length - mth->code_length 
		      + (mth->new_exception_table_length - 
			 mth->exception_table_length)*8 
		      - mth->line_table_size - mth->local_var_table_size);

		pack2(fp+6, mth->new_max_stack); 
		pack2(fp+8, mth->new_max_locals);
 		pack4(fp+10, mth->new_code_length);

		fwrite(fp, 14, 1, f);
		fwrite(mth->new_code, mth->new_code_length, 1, f);
		fp += 14 + mth->code_length;
		pack2(fp, mth->new_exception_table_length);
		fwrite(fp, 2, 1, f);
		fp += 2 + mth->exception_table_length*8;
		fwrite(mth->new_exception_table, 
		       mth->new_exception_table_length*8, 1, f);

		int method_attr_count = unpack2(fp); 
		pack2(fp, method_attr_count 
		          - (mth->line_table_size != 0) 
		          - (mth->local_var_table_size != 0));
		fwrite(fp, 2, 1, f);
		fp += 2;	

		while (--method_attr_count >= 0) { 
		    int mth_attr_name = unpack2(fp);
		    int mth_attr_len = unpack4(fp+2);
		    attr = (const_utf8*)constant_pool[mth_attr_name];
		    if (attr->str != "LineNumberTable" &&
			attr->str != "LocalVariableTable") 
		    { 
			fwrite(fp, mth_attr_len+6, 1, f);
		    }
		    fp += mth_attr_len + 6;
		}
	    } else { 
		fwrite(fp, attr_len+6, 1, f);
		fp += attr_len + 6;
	    }
	}
    }
    
    int attr_count = unpack2(fp); 
    fwrite(fp, 2, 1, f);
    fp += 2;
    while (--attr_count >= 0) { 
	int attr_len = unpack4(fp+2);
	fwrite(fp, attr_len+6, 1, f);
	fp += attr_len + 6;
    }
    fclose(f);
}



class_method::class_method(utf_string const& mth_name, 
			   utf_string const& mth_desc,
			   int mth_attr, class_file* cls) 
: name(mth_name), desc(mth_desc), attr(mth_attr)
{
    this_class = cls;
    line_table_size = 0;
    local_var_table_size = 0;
    recursive = false;
    call_count = 0;

    byte type[MAX_METHOD_PARAMS];
    type[0] = t_object;
    int n_params = (mth_attr & f_static) ? 0 : 1;
    const byte* p = desc.data;
    assert(*p == '(');
    while (*++p != ')') { 
	switch (*p) { 
	  case '[':
	    while (*++p == '[');
	    if (*p == 'L') { 
		while (*++p != ';');
	    }
	    type[n_params++] = t_object;
	    break;
	  case 'J':
	    type[n_params++] = t_long;
	    type[n_params++] = t_long;
	    break;
	  case 'D':
	    type[n_params++] = t_double;
	    type[n_params++] = t_double;
	    break;
	  case 'L':
	    while (*++p != ';'); 
	    type[n_params++] = t_object;
	    break;
	  case 'F':
	    type[n_params++] = t_float;
	    break;
	  default:
	    type[n_params++] = t_int;
	}
    }
    assert(n_params < MAX_METHOD_PARAMS);
    n_parameters = n_params;
    param_type = new byte[n_params];
    memcpy(param_type, type, n_params);
}


void class_method::optimize(generation_context& ctx, bool inline_function) 
{ 
    int len;
    int pc;
    int i;
    
    int code_length = this->code_length;

    if (code_length == 0 || size_t(code_length) >= sizeof(ctx.code_buffer)) { 
	this->new_code = this->code;
	this->new_code_length = code_length;
	return;
    }

    byte* code = this->code;
    byte* new_code = ctx.code_buffer;
    byte* ip = ctx.code_ptr;

    
    int sp = 0;
    struct { 
	int   index;
	byte* ip; 
    } stack[MAX_METHOD_PARAMS];

    int* variable_mapping = new int[max_locals];

    int* constant_mapping = this_class->constant_mapping;
    if (ctx.this_class != this_class->mapping_for_class) { 
	memset(constant_mapping, 0, 
	       sizeof(int)*this_class->constant_pool_count);
	this_class->mapping_for_class = ctx.this_class;
    }
    int* relocation_table = new int[code_length+1];
    relocation_record* relocation_chain = NULL;

    constant** constant_pool = this_class->constant_pool;

    assert (n_parameters <= max_locals);
    for (i = 0; i < n_parameters - ctx.mapped_variables; i++) { 
	variable_mapping[i] = ctx.max_locals + i;
    } 
    for (; i < n_parameters; i++) {
	variable_mapping[i] = ctx.variable_mapping[i];
    }
    for (; i < max_locals; i++) { 
	variable_mapping[i] = ctx.max_locals + i - ctx.mapped_variables;
    } 
	
    int n_locals = ctx.max_locals + max_locals - ctx.mapped_variables;
    int stack_size = ctx.max_stack + max_stack;
    int max_local_vars = n_locals;
    int max_stack_size = stack_size;
    recursive = true;

    for (pc = 0; pc < code_length;) { 
	int new_pc = ip - new_code;
	assert((size_t)new_pc < sizeof(ctx.code_buffer));
	relocation_table[pc] = new_pc;
	if (relocation_record::check(relocation_chain, pc, new_pc)) { 
	    sp = 0; // entry poiny
	}
	switch (code[pc]) { 
	  case jsr:
	  case ifeq:
	  case ifne:
	  case iflt:
	  case ifge:
	  case ifgt:
	  case ifle:
	  case if_icmpeq:
	  case if_icmpne:
	  case if_icmplt:
	  case if_icmpge:
	  case if_icmpgt:
	  case if_icmple:
	  case if_acmpeq:
	  case if_acmpne:
	  case ifnull:
	  case ifnonnull:
	  case goto_near:
	    ip[0] = code[pc]; 
	    ip[1] = code[pc+1];
	    ip[2] = code[pc+2];
	    relocation_record::add(relocation_chain, pc, new_pc,
				   ip+1, 2, relocation_table);
	    ip += 3;
	    pc += 3;
	    break;
	  case jsr_w:
	  case goto_w:
	    ip[0] = code[pc]; 
	    ip[1] = code[pc+1];
	    ip[2] = code[pc+2];
	    ip[3] = code[pc+3];
	    ip[4] = code[pc+4];
	    relocation_record::add(relocation_chain, pc, new_pc,
				   ip+1, 4, relocation_table);
	    ip += 5;
	    pc += 5;
	    break;
	  case tableswitch:
	    { 
		int old_pc = pc;
		ip[0] = code[pc];
		pc += 4 - (old_pc & 3);
		ip += 4 - (new_pc & 3);
		int low = unpack4(&code[pc+4]); 
		int high = unpack4(&code[pc+8]);
		memcpy(ip, code+pc, 12 + 4*(high-low+1));
		relocation_record::add(relocation_chain, old_pc, new_pc,
				       ip, 4, relocation_table);
		ip += 12;
		pc += 12 + 4*(high-low+1);
		while (low <= high) { 
		    relocation_record::add(relocation_chain, old_pc, new_pc, 
					   ip, 4, relocation_table);
		    ip += 4;
		    low += 1;
		}
		break;
	    }
	  case lookupswitch:
	    { 
		int old_pc = pc;
		ip[0] = code[pc];
		pc += 4 - (old_pc & 3);
		ip += 4 - (new_pc & 3);
		int n_pairs = unpack4(&code[pc+4]); 
		memcpy(ip, code+pc, 8 + 8*n_pairs);
		relocation_record::add(relocation_chain, old_pc, new_pc,
				       ip, 4, relocation_table);
		ip += 8;
		pc += 8 + 8*n_pairs;
		while (--n_pairs >= 0) { 
		    relocation_record::add(relocation_chain, old_pc, new_pc, 
					   ip+4, 4, relocation_table);
		    ip += 8;
		}
		break;
	    }
	  case iload_0:
	  case iload_1:
	  case iload_2:
	  case iload_3:
	    i = variable_mapping[code[pc] - iload_0];
	    stack[sp].index = i;
	    stack[sp++].ip = ip;
	    if (i <= 3) { 
		*ip++ = iload_0 + i;
	    } else if (i < 256) { 
		*ip++ = iload;
		*ip++ = i;
	    } else { 
		*ip++ = wide;
		*ip++ = iload;
		ip = pack2(ip, i);
	    }
	    pc += 1;
	    continue;
	  case iload:
	  case aload:
	  case fload:
	    i = variable_mapping[code[pc+1]];
	    stack[sp].index = i;
	    stack[sp++].ip = ip;
	    if (i < 256) { 
		*ip++ = code[pc];
		*ip++ = i;
	    } else { 
		*ip++ = wide;
		*ip++ = code[pc];
		ip = pack2(ip, i);
	    } 
	    pc += 2;
	    continue;
	  case aload_0:
	  case aload_1:
	  case aload_2:
	  case aload_3:
	    i = variable_mapping[code[pc] - aload_0];
	    stack[sp].index = i;
	    stack[sp++].ip = ip;
	    if (i <= 3) { 
		*ip++ = aload_0 + i;
	    } else if (i < 256) { 
		*ip++ = aload;
		*ip++ = i;
	    } else { 
		*ip++ = wide;
		*ip++ = aload;
		ip = pack2(ip, i);
	    }
	    pc += 1;
	    continue;
	  case fload_0:
	  case fload_1:
	  case fload_2:
	  case fload_3:
	    i = variable_mapping[code[pc] - fload_0];
	    stack[sp].index = i;
	    stack[sp++].ip = ip;
	    if (i <= 3) { 
		*ip++ = fload_0 + i;
	    } else if (i < 256) { 
		*ip++ = fload;
		*ip++ = i;
	    } else { 
		*ip++ = wide;
		*ip++ = fload;
		ip = pack2(ip, i);
	    }
	    pc += 1;
	    continue;
	  case lload_0:
	  case lload_1:
	  case lload_2:
	  case lload_3:
	    i = variable_mapping[code[pc] - lload_0];
	    stack[sp].index = i;
	    stack[sp++].ip = ip;
	    stack[sp].index = i + 1;
	    stack[sp++].ip = ip;
	    if (i <= 2) { 
		*ip++ = lload_0 + i;
	    } else if (i < 255) { 
		*ip++ = lload;
		*ip++ = i;
	    } else { 
		*ip++ = wide;
		*ip++ = lload;
		ip = pack2(ip, i);
	    }
	    pc += 1;
	    continue;
	  case lload:
	  case dload:
	    i = variable_mapping[code[pc+1]];
	    stack[sp].index = i;
	    stack[sp++].ip = ip;
	    stack[sp].index = i + 1;
	    stack[sp++].ip = ip;
	    if (i < 255) { 
		*ip++ = code[pc];
		*ip++ = i;
	    } else { 
		*ip++ = wide;
		*ip++ = code[pc];
		ip = pack2(ip, i);
	    } 
	    pc += 2;
	    continue;
	  case dload_0:
	  case dload_1:
	  case dload_2:
	  case dload_3:
	    i = variable_mapping[code[pc] - dload_0];
	    stack[sp].index = i;
	    stack[sp++].ip = ip;
	    stack[sp].index = i + 1;
	    stack[sp++].ip = ip;
	    if (i <= 2) { 
		*ip++ = dload_0 + i;
	    } else if (i < 255) { 
		*ip++ = dload;
		*ip++ = i;
	    } else { 
		*ip++ = wide;
		*ip++ = dload;
		ip = pack2(ip, i);
	    }
	    pc += 1;
	    continue;
	  case istore_0:
	  case istore_1:
	  case istore_2:
	  case istore_3:
	    i = variable_mapping[code[pc] - istore_0];
	    if (i <= 3) { 
		*ip++ = istore_0 + i;
	    } else if (i < 256) { 
		*ip++ = istore;
		*ip++ = i;
	    } else { 
		*ip++ = wide;
		*ip++ = istore;
		ip = pack2(ip, i);
	    }
	    pc += 1;
	    break;
	  case istore:
	  case astore:
	  case fstore:
	    i = variable_mapping[code[pc+1]];
	    if (i < 256) { 
		*ip++ = code[pc];
		*ip++ = i;
	    } else { 
		*ip++ = wide;
		*ip++ = code[pc];
		ip = pack2(ip, i);
	    } 
	    pc += 2;
	    break;
	  case astore_0:
	  case astore_1:
	  case astore_2:
	  case astore_3:
	    i = variable_mapping[code[pc] - astore_0];
	    if (i <= 3) { 
		*ip++ = astore_0 + i;
	    } else if (i < 256) { 
		*ip++ = astore;
		*ip++ = i;
	    } else { 
		*ip++ = wide;
		*ip++ = astore;
		ip = pack2(ip, i);
	    }
	    pc += 1;
	    break;
	  case fstore_0:
	  case fstore_1:
	  case fstore_2:
	  case fstore_3:
	    i = variable_mapping[code[pc] - fstore_0];
	    if (i <= 3) { 
		*ip++ = fstore_0 + i;
	    } else if (i < 256) { 
		*ip++ = fstore;
		*ip++ = i;
	    } else { 
		*ip++ = wide;
		*ip++ = fstore;
		ip = pack2(ip, i);
	    }
	    pc += 1;
	    break;
	  case lstore_0:
	  case lstore_1:
	  case lstore_2:
	  case lstore_3:
	    i = variable_mapping[code[pc] - lstore_0];
	    if (i <= 2) { 
		*ip++ = lstore_0 + i;
	    } else if (i < 255) { 
		*ip++ = lstore;
		*ip++ = i;
	    } else { 
		*ip++ = wide;
		*ip++ = lstore;
		ip = pack2(ip, i);
	    }
	    pc += 1;
	    break;
	  case lstore:
	  case dstore:
	    i = variable_mapping[code[pc+1]];
	    if (i < 255) { 
		*ip++ = code[pc];
		*ip++ = i;
	    } else { 
		*ip++ = wide;
		*ip++ = code[pc];
		ip = pack2(ip, i);
	    } 
	    pc += 2;
	    break;
	  case dstore_0:
	  case dstore_1:
	  case dstore_2:
	  case dstore_3:
	    i = variable_mapping[code[pc] - dstore_0];
	    if (i <= 2) { 
		*ip++ = dstore_0 + i;
	    } else if (i < 255) { 
		*ip++ = dstore;
		*ip++ = i;
	    } else { 
		*ip++ = wide;
		*ip++ = dstore;
		ip = pack2(ip, i);
	    }
	    pc += 1;
	    break;
	  case ireturn:
	  case freturn:
	  case areturn:
	  case dreturn:
	  case lreturn:
	  case vreturn:
	    if (inline_function) { 
		if (pc + 1 != code_length) { 
		    *ip++ = goto_near;
		    pack2(ip, code_length - pc);
		    relocation_record::add(relocation_chain, pc, new_pc, 
					   ip, 2, relocation_table);
		    ip += 2;
		}
		pc += 1;
	    } else {
		*ip++ = code[pc++];
	    }
	    break;
	  case ldc:
	    i = code[pc+1];
	    if (this_class != ctx.this_class) { 
		i = constant::relocate(i,ctx, constant_pool, constant_mapping);
	    }
	    pc += 2;
	    if (i < 256) { 
		*ip++ = ldc;
		*ip++ = i;
	    } else { 
		*ip++ = ldc_w;
		ip = pack2(ip, i);
	    }
	    break;
	  case getfield:
	  case getstatic:
	  case putstatic:
	  case putfield:
	  case ldc_w:
	  case ldc2_w:
	  case instanceof:
	  case checkcast:
	  case anewarray:
	  case anew:
	    *ip++ = code[pc];
	    i = unpack2(code+pc+1);
	    if (this_class != ctx.this_class) { 
		i = constant::relocate(i,ctx, constant_pool, constant_mapping);
	    }
	    ip = pack2(ip, i);
	    pc += 3;
	    break;
	  case multianewarray:
	    *ip++ = code[pc];
	    i = unpack2(code+pc+1);
	    if (this_class != ctx.this_class) { 
		i = constant::relocate(i,ctx, constant_pool, constant_mapping);
	    }
	    ip = pack2(ip, i);
	    *ip++ = code[pc+3];
	    pc += 4;
	    break;
	  case invokeinterface:
	    *ip++ = code[pc];
	    i = unpack2(code+pc+1);
	    if (this_class != ctx.this_class) { 
		i = constant::relocate(i,ctx, constant_pool, constant_mapping);
	    }
	    ip = pack2(ip, i);
	    *ip++ = code[pc+3];
	    *ip++ = code[pc+4];
	    pc += 5;
	    break;
	  case invokespecial:	
	  case invokevirtual:	
	  case invokestatic:
	    {
		i = unpack2(code+pc+1);
		const_ref* cref = (const_ref*)constant_pool[i];
		const_name_and_type* nm = 
		    (const_name_and_type*)constant_pool[cref->name_and_type];
		const_utf8* name = (const_utf8*)constant_pool[nm->name];
		const_utf8* type = (const_utf8*)constant_pool[nm->desc];
		const_class* cc = (const_class*)constant_pool[cref->cls];
		const_utf8* class_name = (const_utf8*)constant_pool[cc->name];
		class_file* cls = class_file::find(class_name->str);
		if (this_class != ctx.this_class) { 
		    i = constant::relocate(i, ctx, constant_pool, 
					   constant_mapping);
		}

		// Jacob
		class_method* mth = NULL;
		if( cls != NULL ) {
			mth = cls->find_method(name->str, type->str);
			while( mth == NULL ) {
				cls = class_file::find(*(cls->super_class_name));
				if( cls == NULL ) {
					break;
				}
				mth = cls->find_method(name->str, type->str);
			}
		}
		//

		if (cls != NULL
		    && (!not_inline_constructor || name->str.data[0] != '<')
		    && (!not_inline_virtual || code[pc] != invokevirtual)
		    && (code[pc] != invokevirtual || cls->final))
		{ 
		    //class_method* mth = 
			//cls->find_method(name->str, type->str);
		    assert(mth != NULL);
		    if (mth->code_length <= inline_threshold
			&& !mth->recursive
                        && mth->exception_table_length == 0
			&& !(mth->attr & (f_abstract|f_native))
			&& mth->call_count >= min_call_count) 
		    { 
			int n = mth->n_parameters;
			ctx.mapped_variables = sp > n ? n : sp;
			while (--sp >= 0 && n > 0) { 
			    ctx.variable_mapping[--n] = stack[sp].index;
			    ip = stack[sp].ip;
			}
			while (--n >= 0) { 
			    int cop = 0, wide_cop = 0;
			    switch (mth->param_type[n]) {
			      case t_object:
				cop = astore_0;
				wide_cop = astore;
				break;
			      case t_int:
				cop = istore_0;
				wide_cop = istore;
				break;
			      case t_double:
				cop = dstore_0;
				wide_cop = dstore;
				n -= 1;
				assert(n >= 0);
				break;
			      case t_float:
				cop = fstore_0;
				wide_cop = fstore;
				break;
			      case t_long:
				cop = lstore_0;
				wide_cop = lstore;
				n -= 1;
				assert(n >= 0);
				break;
			      default:
				assert(0);
			    }
			    int param = n + n_locals;
			    if (param <= 3) { 
				*ip++ = cop + param;
			    } else { 
				*ip++ = wide_cop;
				*ip++ = param;
			    }
			}
			ctx.code_ptr = ip;
			ctx.max_stack = stack_size;
			ctx.max_locals = n_locals;
#if 0
printf("Inline method %.*s.%.*s%.*s  in  %.*s.%.*s%.*s\n",  cls->class_name->len, cls->class_name->data, mth->name.len, mth->name.data, mth->desc.len, mth->desc.data, this_class->class_name->len, this_class->class_name->data, this->name.len, this->name.data, this->desc.len, this->desc.data);
#endif
			mth->optimize(ctx, true);
			if (ctx.max_locals > max_local_vars) { 
			    max_local_vars = ctx.max_locals;
			}
			if (ctx.max_stack > max_stack_size) { 
			    max_stack_size = ctx.max_stack;
			}
			ip = ctx.code_ptr;
			n_inline_calls += 1;
			pc += 3;
			break;
		    } else if (code[pc] == invokevirtual 
			       && !not_use_invokespecial) 
		    { 
			n_nonvirtual_calls += 1;
			*ip++ = invokespecial;
			ip = pack2(ip, i);
			pc += 3;
			break;
		    }
		}
		*ip++ = code[pc];
		ip = pack2(ip, i);
		pc += 3;
		break;
	    }
	  case iinc:
	    i = variable_mapping[code[pc+1]];
	    if (i < 256) { 
		*ip++ = iinc;
		*ip++ = i;
		*ip++ = code[pc+2];
	    } else { 
		*ip++ = wide;
		*ip++ = iinc;
		ip = pack2(ip, i);
		ip = pack2(ip, (signed char)code[pc+2]);
	    } 
	    pc += 3;
	    break;
	  case wide:
	    *ip++ = code[pc++];
	    switch(*ip++ = code[pc++]) { 
	      case iload:
	      case aload:
	      case fload:
		i = variable_mapping[unpack2(code+pc)];
		stack[sp].index = i;
		stack[sp++].ip = ip - 2;
		ip = pack2(ip, i);
		pc += 2;
		continue;
	      case dload:
	      case lload:
		i = variable_mapping[unpack2(code+pc)];
		stack[sp].index = i;
		stack[sp++].ip = ip - 2;
		stack[sp].index = i + 1;
		stack[sp++].ip = ip + 2;
		ip = pack2(ip, i);
		pc += 2;
		continue;
	      case istore:
	      case astore:
	      case fstore:
	      case dstore:
	      case lstore:
		ip = pack2(ip, variable_mapping[unpack2(code+pc)]);
		pc += 2;
		break;
	      case iinc:
		ip = pack2(ip, variable_mapping[unpack2(code+pc)]);
		*ip++ = code[pc+2];
		*ip++ = code[pc+3];
		pc += 4;
		break;
	    }
	    break;
	  default:
	    len = vbm_instruction_length[code[pc]];
	    memcpy(ip, code+pc, len);
	    pc += len;
	    ip += len;
	    break;
	}
	sp = 0;
    }
    relocation_record::check(relocation_chain, pc, ip - new_code);
    assert(relocation_chain == NULL);

    ctx.code_ptr = ip;
    ctx.max_locals = max_local_vars;
    ctx.max_stack = max_stack_size;

    relocation_table[code_length] = ip - new_code;

    byte* p = ctx.extend_exception_buffer(exception_table_length*8);
    for (i = 0; i < exception_table_length; i++) { 
	p = pack2(p, relocation_table[unpack2(exception_table + i*8)]);
        p = pack2(p, relocation_table[unpack2(exception_table + i*8 + 2)]);
	p = pack2(p, relocation_table[unpack2(exception_table + i*8 + 4)]);
	*p++ = exception_table[i*8 + 6];
	*p++ = exception_table[i*8 + 7];
    }
    delete[] variable_mapping;
    delete[] relocation_table;

    if (!inline_function) { 
	new_code_length = ip - new_code;
	this->new_code = new byte[new_code_length];
	memcpy(this->new_code, new_code, new_code_length);

	new_exception_table_length = 
	    (ctx.exception_ptr - ctx.exception_buffer) / 8;
	new_exception_table = new byte[new_exception_table_length*8];
	memcpy(new_exception_table, ctx.exception_buffer, 
	       new_exception_table_length*8);

	new_max_stack = max_stack_size;
	new_max_locals = max_local_vars;
    }
    recursive = false;
}


int main(int argc, char* argv[])
{
    int i;
    if (argc == 1) { 
	fprintf(stderr, "JavaGO: Java Global Optimizer    version %3.2f\n"
		"Usage: javago [options] {java-class-files...}\n"
		"where options are:\n"
                "  -not-inline-constructor   don't inline constructors body\n"
                "  -not-inline-virtual       don't inline virtual methods\n"
                "  -not-use-invokespecial    don't replace invokevirtual with invokespecial\n"
		"  -threshold VALUE          don't inline methods with byte code size > VALUE\n"
		"  -profile FILE-NAME COUNT  don't inline methods with call count < COUNT\n", VERSION);
	return EXIT_FAILURE;
    }	
    for (i = 1; i < argc; i++) { 
	char* file_name = argv[i];
	if (*file_name == '-') { 
	    if (strcmp(file_name, "-threshold") == 0) { 
		if (i+1 == argc ||
		    sscanf(argv[++i], "%d", &inline_threshold) != 1) 
		{
		    fprintf(stderr, "Bad value of inline tshreshold option\n");
		} 
	    } else if (strcmp(file_name, "-profile") == 0) { 
		if (i+2 >= argc ||
		    sscanf(argv[i+=2], "%d", &min_call_count) != 1) 
		{
		    fprintf(stderr, "Bad value of count value for profile "
			    "option\n");
		} else { 
		    profile_name = argv[i-1];
		}
	    } else if (strcmp(file_name, "-not-inline-constructor") == 0) { 
		not_inline_constructor = true;
	    } else if (strcmp(file_name, "-not-inline-virtual") == 0) { 
		not_inline_virtual = true;
	    } else if (strcmp(file_name, "-not-use-invokespecial") == 0) { 
		not_use_invokespecial = true;
	    } else { 
		fprintf(stderr, "Unknown option: %s\n", file_name);
	    }
	    continue;
	}
	load_file(file_name);
    }
    class_file::optimize();
    printf("Substitute %d method invocations with inline code\n"
	   "Replace %d virtual calls with non-virtual ones\n", 
	   n_inline_calls, n_nonvirtual_calls);
    return EXIT_SUCCESS;
}
 


